
from src.rest_playbook_micro.assertion import Assertion
from src.rest_playbook_micro._play import Play as P


class TestAssertion(Assertion):

    def __init__(self) -> None:
        pass

    def run_assert(self, play: P, variables) -> bool:

        match play.STEP:
            case '1':
                self.run_01(play, variables)
            case '2':
                self.run_02(play, variables)
            case _:
                pass

        return True

    def run_01(self, play: P, variables) -> bool:
        play.result.assert_test = True
        return True

    def run_02(self, play: P, variables) -> bool:
        play.result.assert_test = False
        return True



from src.rest_playbook_micro.rest_playbook import RESTPlaybook as RP


def test_status():
    rp = RP(
        "tests/test_status/simple.playbook",
        "tests/test_status/simple_test.scenario",
        "tests/test_status/file_test.vars"
    )
    results = rp.run_playbook()
    assert len(results) == 2
    assert results[0].status == 200
    assert rp.playbook.plays[0].valid_status() is True
    assert results[1].status == 404
    assert rp.playbook.plays[1].valid_status() is False



from src.rest_playbook_micro._variables import Variables as VS
from src.rest_playbook_micro._variable import Variable as V


def test_variable():
    v = V(name="NAME", value="VALUE")
    assert v.name == "NAME"


def test_variables():
    v = VS("./tests/file_test.vars")
    assert v.variables[0].name == "PAGE"
    assert v.variables[0].value == "1"
    assert v.variables[2].name == "SERVICE_ID"
    assert v.variables[2].value == "SERVICEORDER001"
    assert v.variables[3].name == "DOUBLE_EQUALS"
    assert v.variables[3].value == "123=456===ABC"



from src.rest_playbook_micro.rest_playbook import RESTPlaybook as RP

def test_no_input():
    rp = RP(
        "",
        "",
        ""
    )
    results = rp.run_playbook()
    assert len(results) == 0

def test_missing_scenario():
    rp = RP(
        "tests/no_scenario.playbook",
        "",
        "tests/file_test.vars"
    )
    results = rp.run_playbook()
    assert len(results) == 3
    assert results[0].body == 'aa - Scenario not found'
    assert results[1].body == 'ab - Scenario not found'
    assert results[2].body == 'ac - Scenario not found'

def test_missing_playbook():
    rp = RP(
        "",
        "tests/simple_test.scenario",
        "tests/file_test.vars"
    )
    results = rp.run_playbook()
    assert len(results) == 0

def test_success_playbook():
    rp = RP(
        "tests/simple.playbook",
        "tests/simple_test.scenario",
        "tests/file_test.vars"
    )
    results = rp.run_playbook()
    assert len(results) == 3
    assert results[0].status == 200
    assert results[0].body == ''
    assert results[1].status == 201
    assert results[1].body == ''
    assert results[2].status == 200
    assert results[2].body.startswith('[{"id": 1, "name": "Company One"}, {"id": 2,')



import time
from ._result import Result as R
from ._scenarios import Scenarios as S
from ._variables import Variables as V


class Play():
    """
    Invidual step in a playbook, does not have direct knowledge of the
    scenario it relates to, but instead a unique reference to a name.
    Manages parsing of the response and time to the next step
    """

    NAME: str
    WAIT: int
    ASSERT: str
    STEP: str
    # For testing purposes we can have the outbound body
    # put back into the play object
    RETURN_BODY: bool
    NEXT_STEP: str
    VALID_STATUS: list
    result: R
    scenario: S

    def __init__(self,
                 name: str,
                 step: str,
                 assert_exec: str,
                 valid_status: list[str],
                 next_step: str = None,
                 wait: int = 30,
                 return_body: bool = False) -> None:
        if name:
            self.NAME = name
        if step:
            self.STEP = step
        if assert_exec:
            self.ASSERT = assert_exec
        if valid_status:
            self.VALID_STATUS = valid_status
        else:
            self.VALID_STATUS = []
        self.NEXT_STEP = next_step
        self.WAIT = wait
        self.RETURN_BODY = return_body

    def run(self, scenarios: S, variables: V) -> R:
        self.result = scenarios.run_scenario(
            self.NAME, variables, return_body=self.RETURN_BODY)
        self.valid_status()
        time.sleep(self.WAIT)
        return self.result

    def valid_status(self) -> None:
        if str(self.result.status) in self.VALID_STATUS:
            self.result.valid = True
        return self.result.valid



"""
Util repository
"""
import logging
import sys
import regex as re


from itertools import groupby


class utils():

    logging.basicConfig(
        format='%(asctime)s | %(levelname)s | %(message)s', level=logging.DEBUG)

    debug = False

    def _debug(self, message):
        if self.debug:
            logging.debug(str(message))

    def parse_vars_file(self, file: str) -> []:
        var_blocks = self._split_var_file(self.load_file(file))
        return list(map(self._split_var_file, var_blocks))

    def _split_var_file(self, e):
        split_row = str(e).split('=', 1)
        if len(split_row) == 2:
            return (f'%ENV_{split_row[0]}%', split_row[1])

    def find_mandatory(self, line: str) -> str:
        if line.startswith("%MANDATORY:") is False:
            return line
        x = re.findall(r"(%MANDATORY:ENV_[A-Z_0-9]+%)", line)
        var_to_find = f'%{str(x[0]).split(":")[1]}'
        filtered_list = filter(lambda c: c[0] == var_to_find, vars)
        filtered_list = list(filtered_list)
        if len(filtered_list) == 0:
            raise ValueError(f"Missing var {var_to_find}")
        return re.sub(r"(%MANDATORY:ENV_[A-Z_0-9]+%)", "", line)

    def find_optional(self, line: str) -> str:
        if line.startswith("%OPTIONAL:") is False:
            return line
        x = re.findall(r"(%OPTIONAL:ENV_[A-Z_0-9]+%)", line)
        var_to_find = f'%{str(x[0]).split(":")[1]}'
        filtered_list = filter(lambda c: c[0] == var_to_find, vars)
        filtered_list = list(filtered_list)
        if len(filtered_list) == 0:
            return ""
        return re.sub(r"(%OPTIONAL:ENV_[A-Z_0-9]+%)", "", line)

    def find_var(self, line: str) -> list:
        """Finds all the ENV variables in a provided string"""
        try:
            x = re.findall(r"(%ENV_[A-Z_0-9]+%)", line)
            return x
        except Exception as err:
            return []

    def set_properties(self, scenario_props: list) -> dict:
        props = {}
        for element in scenario_props:
            e = str(element).split('=', 1)
            props[e[0]] = e[1]
        return props

    def parse_scenario(self, entry: list, a: str, b: str) -> dict:
        properties = [list(g) for k, g in groupby(
            entry, key=lambda x: x != a and x != b) if k]
        # print(properties)
        return properties

    def split_scenario_file(self, file: str) -> list:
        raw_data = self.load_file(file)
        split_scenarios = [list(g) for k, g in groupby(
            raw_data, key=lambda x: x != "---") if k]
        return split_scenarios

    def read_file(self, file_name: str) -> list:
        """
        Uses a filepath + filename string to return a list of all
        the lines in the file, removes newlines in the strings
        """
        try:
            return_array = []
            open_file = file_name
            with open(open_file) as f:
                for line in f:
                    return_array.append(line.strip().replace('\n', ''))
            return return_array
        except Exception as err:
            self._debug(f"Error in readFileNoStrip {err=}, {type(err)=}")
            return []

    def load_file(self, file_name: str) -> list:
        file_data = self.read_file(file_name)
        if file_data is False:
            print('File invalid or not found')
            sys.exit()
        return file_data




import sys

from itertools import groupby
from ._utils import utils
from ._play import Play as P
from ._result import Result as R
from ._scenarios import Scenarios as S
from ._variables import Variables as V
from .assertion import Assertion as A


class Playbook():
    """Parent playbook class, manages each of the steps provided
    """

    plays: list[P]
    u: utils = utils()
    assertion = None

    def __init__(self, playbook_file: str = None) -> None:
        self._parse_playbook_file(playbook_file)
        self.step = None

    def _parse_playbook_file(self, file: str):
        self.plays = self._parse_play_list(file)

    def _parse_play_list(self, file: str):
        playbook_blocks = self._split_playbook_file(file)
        return list(map(self._parse_block, playbook_blocks))

    def _split_playbook_file(self, file: str) -> list:
        raw_data = self.u.load_file(file)
        split_scenarios = [list(g) for k, g in groupby(
            raw_data, key=lambda x: x != "----") if k]
        return split_scenarios

    def _parse_block(self, entry: list):

        props = self.u.set_properties(entry)
        if props.get('NAME', None) is None:
            raise ValueError("Play is missing NAME")
        p = P(
            name=props.get('NAME'),
            wait=int(props.get('WAIT', 30)),
            step=props.get('STEP'),
            assert_exec=props.get('ASSERT'),
            next_step=props.get('NEXT_STEP', None),
            valid_status=str(props.get('VALID_STATUS', '')).split(',')
        )
        return p

    def run(self, scenarios: S, variables: V, assertion: A) -> list[R]:
        ret_val: list[R] = []
        iterator = 0
        for p in self.plays:
            ret_val.append(p.run(scenarios, variables))
            if not p.valid_status:
                return ret_val
            if assertion and not assertion.run_assert(play=p, variables=variables):
                return ret_val
            iterator += 1
            if iterator > 500:
                return ret_val
        return ret_val



"""
Private module providing Scenario class
"""

import regex as re

from rest_client_micro.rest_client import RESTClient as rc
from rest_client_micro.rest_object import RESTObject as ro
from rest_client_micro.response import Response as Resp

from ._result import Result as R
from ._utils import utils
from ._variables import Variables as V


class Scenario():

    properties: dict
    headers: list
    params: list
    body: list
    raw_body: list
    u: utils = utils()

    def __init__(self) -> None:
        self.properties = {}
        self.headers = []
        self.params = []
        self.body = []
        self.raw_body = []

    def run_scenario(self, variables: V, return_body=False) -> R:
        # print(f"Running {self.properties['NAME']}")
        self.apply_variables_body(variables)
        client = rc()
        # client.debug = True
        rest_object = ro()
        rest_object.endpoint = self.properties.get(
            'ENDPOINT', 'localhost:3000')
        rest_object.payload = self.get_raw_body()
        rest_result = client.execute(rest_object)
        if rest_result.error:
            result = R(response=rest_result.error_text)
        else:
            result = R(response=rest_result.response,
                       status=rest_result.status)
        if return_body:
            result.outbound = rest_result.outbound
        return result

    def apply_variables_body(self, v: V):
        for line in self.body:
            new_line = line
            # new_line = self._find_mandatory(new_line, v)
            new_line = self.find_optional(new_line, v)
            if len(new_line) == 0:
                continue
            line_vars = self.u.find_var(line)
            if len(line_vars) == 0:
                self.raw_body.append(new_line)
                continue
            # print(line_vars)
            self.raw_body.append(self._generate_line(new_line, v.variables))

    def _find_mandatory(self, line: str, v: V) -> str:
        if line.startswith("%MANDATORY:") is False:
            return line
        x = re.findall(r"(%MANDATORY:ENV_[A-Z_0-9]+%)", line)
        var_to_find = f'%{str(x[0]).split(":")[1]}'
        filtered_list = filter(
            lambda c: f"%ENV_{c.name}%" == var_to_find, v.variables)
        filtered_list = list(filtered_list)
        if len(filtered_list) == 0:
            raise ValueError(f"Missing var {var_to_find}")
        return re.sub(r"(%MANDATORY:ENV_[A-Z_0-9]+%)", "", line)

    def find_optional(self, line: str, v: V) -> str:
        if line.startswith("%OPTIONAL:") is False:
            return line

        x = re.findall(r"(%OPTIONAL:ENV_[A-Z_0-9]+%)", line)
        var_to_find = f'%{str(x[0]).split(":")[1]}'
        filtered_list = filter(
            lambda c: f"%ENV_{c.name}%" == var_to_find, v.variables)
        filtered_list = list(filtered_list)
        if len(filtered_list) == 0:
            return ""
        return re.sub(r"(%OPTIONAL:ENV_[A-Z_0-9]+%)", "", line)

    def _generate_line(self, line: str, variables: list[V]) -> str:
        line_vars = self.u.find_var(line)
        if len(line_vars) == 0:
            return line
        for v in variables:
            if f"%ENV_{v.name}%" in line_vars:
                line = line.replace(f"%ENV_{v.name}%", v.value)
                break
        for v in line_vars:
            if v in line:
                line = line.replace(v, "")
        return line

    def get_raw_body(self) -> str:
        return "".join(self.raw_body)

    def get_body(self) -> str:
        return "".join(self.body)

    def print(self) -> None:
        print(
            f"""--
            {self.properties}
{self.headers}
{self.params}
{self.get_raw_body()}
{self.get_body()}
--"""
        )



from ._variables import Variables as V
from ._play import Play as P


class Assertion():

    def __init__(self) -> None:
        pass

    def run_assert(self, play: P, variables: V) -> bool:

        match play.STEP:
            case _:
                return self.run_01(play, variables)

    def get_variable(self, variables: V, var_name: str):
        pass

    def run_01(self, play: P, variables: V) -> bool:
        return True



"""
Run dynamic and ordered REST transactions
"""

from .rest_playbook import RESTPlaybook
from .assertion import Assertion
from ._play import Play

VERSION = (0, 1, 2)

VERSION_STRING = ".".join(map(str, VERSION))

RESTPlaybook

Assertion

Play



"""
Module providing RESTClient class
"""
import os

from pathlib import Path
from ._scenarios import Scenarios as S
from ._variables import Variables as V
from ._playbook import Playbook as P
from ._result import Result as R
from .assertion import Assertion as A


class RESTPlaybook():
    """Top level class for running playbooks, manages the orchestration
    of each file
    """
    app_name: str = 'rest_playbook_micro'
    config_dir: str = os.path.join(
        str(Path.home()),
        ".config/",
        app_name
    )

    scenario_file: str = ''
    var_file: str = ''
    playbook_file: str = ''

    scenarios: S
    variables: V
    playbook: P
    assertion: A

    def __init__(self,
                 p_file: str = "env.playbook",
                 s_file: str = "env.scenario",
                 v_file: str = "env.vars",
                 auth_file: str = None,
                 assertion: A = None
                 ) -> None:
        """Create a new Playbook object

        :param p_file: File path + name for playbook file
        :param s_file: File path + name for playbook file
        :param v_file: File path + name for playbook file
        """
        self.set_scenario_file(s_file)
        self.set_var_file(v_file)
        self.set_playbook_file(p_file)
        self.assertion = assertion

    def set_playbook_file(self, playbook_file: str) -> None:
        self.playbook_file = playbook_file
        self.generate_playbook(self.playbook_file)

    def generate_playbook(self, playbook_file: str = None) -> None:
        if playbook_file is None:
            playbook_file = self.playbook_file
        self.playbook = P(playbook_file)

    def set_scenario_file(self, scenario_file: str) -> None:
        self.scenario_file = scenario_file
        self.generate_scenarios(self.scenario_file)

    def set_var_file(self, var_file: str) -> None:
        self.var_file = var_file
        self.generate_variables(self.var_file)

    def generate_variables(self, var_file: str) -> None:
        self.variables = V(var_file)

    def generate_scenarios(self, scenario_file: str) -> None:
        self.scenarios = S(scenario_file)

    def run_scenario(self, scenario_name: str) -> R:
        return self.scenarios.run_scenario(
            scenario_name, self.variables
        )

    def run_playbook(self) -> list[R]:
        """Runs the full playbook

        :returns: List of result objects for each play in the book
            or up to the point a failure occured
        """
        return self.playbook.run(self.scenarios,
                                 self.variables,
                                 self.assertion)

    def print(self) -> None:
        self.scenarios.print()
        self.variables.print()




from ._variable import Variable as V
from._utils import utils

class Variables():

    var_file: str

    u = utils()
    variables: list[V]

    def __init__(self, var_file: str) -> None:
        self.set_variables(var_file)

    def set_variables(self, var_file: str) -> None:
        self.var_file = var_file
        self._parse_var_file(self.var_file)

    def _parse_var_file(self, file: str) -> None:
        var_blocks = self._split_var_file(file)
        self.variables = list(map(self._parse_block, var_blocks))

    def _parse_block(self, block: str):
        block = block.split('=',1)
        v = V()
        v.name = block[0]
        v.value = block[-1]
        return v

    def _split_var_file(self, file) -> list:
        return self.u.load_file(file)

    def print(self) -> None:
        for v in self.variables:
            v.print()



class Variable():

    name: str
    value: str

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)
        #self.name = f"%ENV_{self.name}%"

    def print(self):
        print(f"{self.name} - {self.value}")



from src.rest_playbook_micro.rest_playbook import RESTPlaybook as RP
from tests.test_assert.playbook_assert_file import TestAssertion

def test_playbook_assert():
    """
    Simple test of an assertion class
    class will try and set `assert_test` property on result
    object to True and then False
    """
    rp = RP(
        "tests/test_assert/simple.playbook",
        "tests/test_assert/simple_test.scenario",
        "tests/test_assert/file_test.vars",
        assertion=TestAssertion()
    )
    results = rp.run_playbook()
    assert results[0].assert_test is True
    assert results[1].assert_test is False



from flask import Flask, json, Response, request

companies = [{"id": 1, "name": "Company One"},
             {"id": 2, "name": "Company Two"}]

api = Flask(__name__)


@api.route('/200', methods=['GET'])
def get_200():
    return Response("", status=200, mimetype='application/json')


@api.route('/201', methods=['GET'])
def get_201():
    return Response("", status=201, mimetype='application/json')


@api.route('/companies', methods=['GET'])
def get_companies():
    return json.dumps(companies)

@api.route('/', methods=['GET'])
def get_index():
    print(request.data)
    return Response("", status=200, mimetype='application/json')


def run_server():
    api.run(port=3876)


if __name__ == '__main__':
    api.run(port=3876)



from src.rest_playbook_micro._playbook import Playbook as P


def test_parse_play_list():
    p = P()
    plays = p._parse_play_list("tests/file_test.playbook")
    assert len(plays) == 3
    assert plays[0].NAME == "aa"
    assert plays[1].WAIT == 1
    assert plays[1].STEP == "2"
    assert plays[1].NEXT_STEP is '3'


def test_playbook():
    p = P("tests/file_test.playbook")
    assert len(p.plays) == 3



import uuid
import time

def get_uuid_id():
    return str(uuid.uuid4())


class Result():

    status: int
    body: str
    outbound: str
    timestamp: float
    uuid: str
    valid: bool

    def __init__(self, response: str = '', status: int = 0) -> None:
        self.uuid = get_uuid_id()
        self.timestamp = time.time()
        self.status = status
        self.body = response
        self.valid = False

    def __str__(self):
        return f"{self.status} - {self.body[:100]}"

    def __repr__(self):
        return f"{self.uuid} - {self.body[:100]}"




from itertools import groupby
from ._utils import utils
from ._scenario import Scenario as S
from ._variables import Variables as V
from ._result import Result as R


class Scenarios():

    scenarios: list[S] = []
    u: utils = utils()

    def __init__(self, scenario_file: str) -> None:
        self._parse_scenario_file(scenario_file)

    def _parse_scenario_file(self, file: str):
        scenario_blocks = self._split_scenario_file(file)
        self.scenarios = list(map(self._parse_block, scenario_blocks))

    def _split_scenario_file(self, file: str) -> list:
        raw_data = self.u.load_file(file)
        split_scenarios = [list(g) for k, g in groupby(
            raw_data, key=lambda x: x != "---") if k]
        return split_scenarios

    def _parse_block(self, entry: list):
        s = S()
        s.properties = self._parse_properties(
            self._parse_scenario(
                entry, "SCENARIO=", "=SCENARIO"
            )[0]
        )
        s.headers = self._parse_scenario(entry, "HEADER=", "=HEADER")[1]
        s.params = self._parse_scenario(entry, "PARAMS=", "=PARAMS")[1]
        s.body = self._parse_scenario(entry, "BODY=", "=BODY")[1]
        return s

    def _parse_properties(self, scenario_props: list) -> dict:
        props = {}
        for element in scenario_props:
            e = str(element).split('=', 1)
            props[e[0]] = e[1]
        return props

    def _parse_scenario(self, entry: list, a: str, b: str) -> dict:
        properties = [list(g) for k, g in groupby(
            entry, key=lambda x: x != a and x != b) if k]
        # print(properties)
        return properties

    def print(self):
        for s in self.scenarios:
            s.print()

    def list_scenarios(self) -> list[str]:
        return []

    def run_scenario(self, scenario_name: str, variables: V, return_body=False) -> R:
        for s in self.scenarios:
            if s.properties.get('NAME', None) == scenario_name:
                return s.run_scenario(variables,return_body=return_body)
        return R(response=f"{scenario_name} - Scenario not found")



from src.rest_playbook_micro._scenarios import Scenarios as S
from src.rest_playbook_micro._variables import Variables as V


def test_scenario_mandatory_var():
    s = S("tests/scenario_substitute.scenario")
    v = V("tests/mandatory_test.vars")
    result = s.run_scenario("ae", v, True)
    assert result.body == ''


def test_scenario_optional_var():
    s = S("tests/scenario_substitute.scenario")
    v = V("tests/file_test.vars")
    result = s.run_scenario("ad", v, True)
    assert result.body == ''
    assert result.outbound == '{"order_id":"ORDER001","optional_property":"SERVICEORDER001"}'
    s = S("tests/scenario_substitute.scenario")
    v = V("tests/optional_test.vars")
    result = s.run_scenario("ad", v, True)
    assert result.body == ''
    asset_text = '{"order_id":"ORDER001","optional_property":"SERVICEORDER001","optional_property2":"SERVICEORDER002"}'
    assert result.outbound == asset_text


def test_scenario_missing_var():
    s = S("tests/scenario_substitute.scenario")
    v = V("tests/file_test.vars")
    result = s.run_scenario("ac", v, True)
    assert result.body == ''
    assert result.outbound == '{"order_id":"ORDER001","missing_property":""}'


def test_scenario_substitute():
    s = S("tests/scenario_substitute.scenario")
    v = V("tests/file_test.vars")
    result = s.run_scenario("aa", v, True)
    assert result.body == ''
    assert result.outbound == '{"order_id":"ORDER001"}'
    result = s.run_scenario("ab", v, True)
    assert result.body == ''
    assert result.outbound == '{"order_id":"ORDER001","untouched_property":"123TEST321"}'


def test_error_scenarios():
    s = S("tests/file_test.scenario")
    v = V("tests/file_test.vars")
    result = s.run_scenario("ab", v)
    assert result.body == ''


def test_success_scenarios():
    s = S("tests/file_test.scenario")
    v = V("tests/file_test.vars")
    result = s.run_scenario("ac", v)
    assert result.body is not None
    assert result.body.startswith('[{"id": 1, "name": "Company One"}')
